"""Internal function to extract timeseries with or without multiprocessing"""
import copy, json, math, os, re, warnings
import numpy as np, pandas as pd
from nilearn.maskers import NiftiLabelsMasker
from nilearn.image import index_img, load_img

def _extract_timeseries(subj_id, nifti_files, mask_files, event_files, confound_files, confound_metadata_files,
                        run_list, tr, condition, parcel_approach, signal_clean_info, verbose, flush_print, task_info):

    # Initialize subject dictionary
    subject_timeseries = {subj_id: {}}

    for run in run_list:
        # Initialize flag; This is for flagging runs where the percentage of volumes exceeding the "fd_threshold"
        # is greater than "outlier_percentage"
        flagged = False

        # Due to prior checking in _setup_extraction within TimeseriesExtractor, run_list = [None] is assumed to be a
        # single file without the run- description
        run_id = "run-0" if run is None else run
        run = run if run is not None else ""

        # Get files from specific run
        files = {}
        files["nifti"] = _grab(run, nifti_files)
        files["mask"] = _grab(run, mask_files)
        files["confound"] = _grab(run, confound_files)
        files["confound_meta"] = _grab(run, confound_metadata_files)
        files["event"] = _grab(run, event_files)

        confound_df = pd.read_csv(files["confound"], sep="\t") if signal_clean_info["use_confounds"] else None

        # Base message
        subject_header, sub_message = _header(task_info["session"], task_info["task"],
                                              run_id.split("-")[-1], files["nifti"], subj_id)

        if verbose:
            print(f"{subject_header}" + f"Preparing for timeseries extraction using - [FILE: {files['nifti']}].",
                  flush=flush_print)

        # Initialize variables
        censor_volumes, scan_list = [], []
        dummy_scans, outlier_limit = None, None

        # Check for non-steady_state if requested
        if signal_clean_info["dummy_scans"]:
            dummy_scans = _get_dummy(signal_clean_info["dummy_scans"], confound_df, verbose, subject_header,
                                     flush_print)

        if signal_clean_info["use_confounds"] and signal_clean_info["fd_threshold"]:
            if "framewise_displacement" in confound_df.columns:
                fd_array = confound_df["framewise_displacement"].fillna(0).values
                # Truncate fd_array if dummy scans;
                if dummy_scans: fd_array = fd_array[dummy_scans:]

                # Get censor volumes vector, outlier_limit, and threshold
                censor_volumes, outlier_limit, threshold = _censor(signal_clean_info["fd_threshold"], fd_array)

                if len(censor_volumes) == fd_array.shape[0]:
                    warnings.warn(f"{subject_header}" + "Processing skipped: Timeseries will be empty due to the "
                                  f"framewise displacement of all volumes exceeding {threshold}.")
                    continue

                if outlier_limit and condition is None:
                    percentage, flagged = _mark(len(censor_volumes), len(fd_array), outlier_limit)

            else:
                warnings.warn(f"{subject_header}" + "`fd_threshold` specified but 'framewise_displacement' column not "
                              "in the confound tsv file. Removal of volumes after nuisance regression will not be done."
                              )

        if files['event']:
            event_df = pd.read_csv(files['event'], sep="\t")
            # Get specific timing information for specific condition
            condition_df = event_df[event_df["trial_type"] == condition]

            if condition_df.empty:
                warnings.warn(f"{subject_header}" + f"[CONDITION: {condition}] - Processing skipped: Condition does "
                              "not exist in the 'trial_type' column of the event file.")
                continue

            # Get condition indices
            scan_list = _get_condition(condition_df, tr, scan_list)

            # Adjust for dummy before censoring since censoring is dummy adjusted above
            if dummy_scans: scan_list = [scan - dummy_scans for scan in scan_list if scan not in range(0, dummy_scans)]

            if censor_volumes:
                scan_list, n = _filter_condition(scan_list, fd_array, condition, subject_header, censor_volumes)
                if outlier_limit: percentage, flagged = _mark(n - len(scan_list), n, outlier_limit)

            if not scan_list:
                warnings.warn(f"{subject_header}" + f"[CONDITION: {condition}] - Processing skipped: Timeseries will "
                              "be empty when filtered to only include volumes from this specific condition. Possibly "
                              "due to TRs corresponding to the condition being removed by `dummy_scans` or filtered "
                              "due to exceeding threshold for `fd_threshold`.")
                continue

        if flagged:
            percentage_message = f"Percentage of volumes exceeding the threshold limit is {percentage*100}%"
            if condition: percentage_message = f"{percentage_message} for [CONDITION: {condition}]"
            warnings.warn(f"{subject_header}" + f"Processing skipped: Run flagged due to more than {outlier_limit*100}% "
                          f"of the volumes exceeding the framewise displacement (FD) threshold limit of {threshold}. "
                          f"{percentage_message}.")
            continue

        timeseries = _continue_extraction(files, confound_df=confound_df, condition=condition,
                                          signal_clean_info=signal_clean_info, dummy_scans=dummy_scans, tr=tr,
                                          parcel_approach=parcel_approach, flush=flush_print, verbose=verbose,
                                          censor_volumes=censor_volumes, scan_list=scan_list,
                                          subject_header=subject_header)

        if timeseries.shape[0] == 0:
            warnings.warn(f"{subject_header}" + f"Timeseries is empty for {run} and will not be appended to the "
                          "`subject_timeseries` dictionary.")
        else:
            subject_timeseries[subj_id].update({run_id: timeseries})

    if not subject_timeseries[subj_id]:
        warnings.warn(f"{sub_message.split(' | RUN:')[0]}]\n"
                      f"{'-'*len(sub_message.split(' | RUN:')[0])}\n"
                      "Processing skipped: No runs were extracted.")
        subject_timeseries = None

    return subject_timeseries

def _continue_extraction(files, confound_df, signal_clean_info, dummy_scans, condition, tr, parcel_approach,
                         flush, verbose, censor_volumes, scan_list, subject_header):
    # Extract confound information of interest and ensure confound file does not contain NAs
    if signal_clean_info["use_confounds"]:

        confound_names = copy.deepcopy(signal_clean_info["confound_names"])

        # Extract first "n" numbers of specified WM and CSF components
        if files["confound_meta"]:
            confound_names = _acompcor(confound_names,
                                       files["confound_meta"],
                                       signal_clean_info["n_acompcor_separate"])

        # Get confounds
        confounds = _get_confounds(confound_names, confound_df, verbose, subject_header, flush)

    # Create the masker for extracting time series
    masker = NiftiLabelsMasker(
        mask_img=files["mask"],
        labels_img=parcel_approach[list(parcel_approach)[0]]["maps"],
        resampling_target='data',
        t_r=tr,
        **signal_clean_info["masker_init"]
    )

    # Load and discard volumes if needed
    nifti_img = load_img(files["nifti"])
    if dummy_scans:
        nifti_img = index_img(nifti_img, slice(dummy_scans, None))
        if signal_clean_info["use_confounds"]:
            confounds.drop(list(range(0, dummy_scans)), axis=0, inplace=True)

    # Extract timeseries
    if signal_clean_info["use_confounds"]: timeseries = masker.fit_transform(nifti_img, confounds=confounds)
    else: timeseries = masker.fit_transform(nifti_img)

    if censor_volumes and not condition: timeseries = np.delete(timeseries, censor_volumes, axis=0)

    if condition:
        # If any out of bound indices, remove them instead of getting indexing error.
        if max(scan_list) > timeseries.shape[0] - 1:
            scan_list = _check_indices(scan_list,timeseries.shape[0],condition,subject_header)
        # Extract condition
        timeseries = timeseries[scan_list,:]

    return timeseries

# Grab files
def _grab(run, files):
    if files: return [file for file in files if run in os.path.basename(file)][0]
    else: return None

# Create message header
def _header(session, task, run_message, nifti_file, subj_id):
    if session:
        session_message = session
    else:
        base_filename = os.path.basename(nifti_file)
        session_id = re.search("ses-(\\S+?)[-_]", base_filename)[0][:-1] if "ses-" in base_filename else None
        session_message = session_id.split("-")[-1] if session_id else None
    sub_message = f'[SUBJECT: {subj_id} | SESSION: {session_message} | TASK: {task} | RUN: {run_message}]'
    underline = '-'*len(sub_message)

    return f"\n{sub_message}\n{underline}\n", sub_message

# Get dummy scan number
def _get_dummy(info, confound_df, verbose, subject_header, flush):
    if isinstance(info, dict) and info["auto"] is True:
        n, flag = len([col for col in confound_df.columns if "non_steady_state" in col]), "auto"
        if "min" in info and n < info["min"]: n, flag = info["min"], "min"
        if "max" in info and n > info["max"]: n, flag = info["max"], "max"
        if n == 0: n = None

        if verbose:
            if flag == "auto":
                if n:
                    print(f"{subject_header}" + "Number of dummy scans to be removed based on "
                            f"'non_steady_state_outlier_XX' columns is {n}.", flush=flush)
                else:
                    print(f"{subject_header}" + "No 'non_steady_state_outlier_XX' columns were found so 0 "
                            "dummy scans will be removed.", flush=flush)
            else:
                print(f"{subject_header}" + f"Default dummy scans set by '{flag}' will be used: {n}",
                        flush=flush)

    return n if isinstance(info, dict) else info

# Create censor vector
def _censor(info, fd_array):
    outlier_limit, threshold = None, None

    if isinstance(info, dict):
        threshold = info["threshold"]
        if "outlier_percentage" in info: outlier_limit = info["outlier_percentage"]
    else:
        threshold = info

    censor_volumes = list(np.where(fd_array > threshold)[0])

    return censor_volumes, outlier_limit, threshold

# Determine if run is flagged; when condition is not specified
def _mark(n_censor, n, outlier_limit):
    percentage = n_censor/n
    flagged = True if percentage > outlier_limit else False

    return percentage, flagged

# Get event condition
def _get_condition(condition_df, tr, scan_list):
    # Convert times into scan numbers to obtain the scans taken when the participant was exposed to the
    # condition of interest; include partial scans
    for i in condition_df.index:
        onset_scan = int(condition_df.loc[i,"onset"]/tr)
        end_scan = math.ceil((condition_df.loc[i,"onset"] + condition_df.loc[i,"duration"])/tr)
        # Add one since range is not inclusive
        scan_list.extend(list(range(onset_scan, end_scan + 1)))

    # Get unique scans to not duplicate information
    scan_list = sorted(list(set(scan_list)))

    return scan_list

def _filter_condition(scan_list, fd_array, condition, subject_header, censor_volumes):
    # Assess if any condition indices greater than fd array to not dilute outlier calculation
    if max(scan_list) > fd_array.shape[0] - 1:
        scan_list = _check_indices(scan_list, fd_array.shape[0], condition, subject_header)
    # Get length of scan list prior to assess outliers if requested
    before_censor = len(scan_list)
    scan_list = [volume for volume in scan_list if volume not in censor_volumes]

    return scan_list, before_censor

# Extract first "n" numbers of specified WM and CSF components
def _acompcor(confound_names, confound_metadata_file, n):
    with open(confound_metadata_file, "r") as confounds_json:
        confound_metadata = json.load(confounds_json)

    acompcors = sorted([acompcor for acompcor in confound_metadata if "a_comp_cor" in acompcor])

    CSF = [CSF for CSF in acompcors if confound_metadata[CSF]["Mask"] == "CSF"][0:n]
    WM = [WM for WM in acompcors if confound_metadata[WM]["Mask"] == "WM"][0:n]

    confound_names.extend(CSF + WM)

    return confound_names

# Get confounds
def _get_confounds(confound_names, confound_df, verbose, subject_header, flush):
    valid_confounds = []
    invalid_confounds = []
    for confound_name in confound_names:
        if "*" in confound_name:
            prefix = confound_name.split("*")[0]
            confounds_list = [col for col in confound_df.columns if col.startswith(prefix)]
        else:
            confounds_list = [col for col in confound_df.columns if col == confound_name]

        if confounds_list: valid_confounds.extend(confounds_list)
        else: invalid_confounds.extend([confound_name])

    if valid_confounds: confounds = confound_df[valid_confounds].fillna(0)
    else: confounds = None

    if invalid_confounds and verbose:
        print(f"{subject_header}" + f"The following confounds were not found: {', '.join(invalid_confounds)}.",
              flush=flush)

    if not confounds.empty and verbose:
        print(f"{subject_header}" + "The following confounds will be for nuisance regression: "
              f"{', '.join(list(confounds.columns))}.", flush=flush)

    return confounds

# Check if indices valid
def _check_indices(scan_list, arr_shape, condition, subject_header):
    warnings.warn(f"{subject_header}" + f"Max scan index for [CONDITION: {condition}] exceeds timeseries max index. "
                  f"Max condition index is {max(scan_list)}, while timeseries max index is {arr_shape - 1}. Timing may "
                  "be misaligned or repetition time incorrect. If intentional, ignore warning. Extracting indices for "
                  "condition only within timeseries range.")

    return [scan for scan in scan_list if scan in range(0,arr_shape)]



from typing import Optional
import os
import matplotlib.pyplot as plt, pandas as pd
from .._utils import (_check_kwargs, _create_display, _save_contents)

def transition_matrix(trans_dict: dict[str, pd.DataFrame], output_dir: Optional[os.PathLike]=None,
                      suffix_title: Optional[str]=None, show_figs: bool = True, save_plots: bool=True,
                      return_df: bool = True,  save_df: bool=True, **kwargs):
    """
    **Generate and Visualize the Averaged Transition Probabilities**

    Uses the "transition_probability" output from ``CAP.calculate_metrics`` to generate and visualize the averaged
    transition probability matrix for all groups from the analysis.

    .. versionadded:: 0.16.2

    Parameters
    ----------
    trans_dict: :obj: `dict[str, pd.DataFrame]`
        A dictionary where the keys are the group names and the values are the pandas DataFrame containing the
        transition probabilities for each subject. This assumes the output from ``CAP.calculate_metrics`` is being used,
        specifically ``metrics_output["transition_probability"]``.

    output_dir : :obj:`os.PathLike` or :obj:`None`, default=None
        Directory to save plots and transition probability matrices DataFrames to. The directory will be created if it
        does not exist. If None, plots and dataFrame will not be saved.

    suffix_title : :obj:`str` or :obj:`None`, default=None
        Appended to the title of each plot as well as the name of the saved file if ``output_dir``
        is provided.

    show_figs : :obj:`bool`, default=True
        Whether to display figures.

    save_plots : :obj:`bool`, default=True
        If True, plots are saves as png images. For this to be used, ``output_dir`` must be specified.

    return_df : :obj:`bool`, default=False
        If True, returns a dictionary with a transition probability matrix for each group.

    save_df : :obj:`bool`, default=False,
        If True, saves the transition probability matrix contained in the DataFrames as csv files. For this to be used,
        ``output_dir`` must be specified.

    kwargs : :obj:`dict`
        Keyword arguments used when modifying figures. Valid keywords include:

        - dpi : :obj:`int`, default=300
            Dots per inch for the figure. Default is 300 if ``output_dir`` is provided and ``dpi`` is not
            specified.
        - figsize : :obj:`tuple`, default=(8, 6)
            Size of the figure in inches.
        - fontsize : :obj:`int`, default=14
            Font size for the plot title, x-axis title, and y-axis title of each plot.
        - xticklabels_size : :obj:`int`, default=8
            Font size for x-axis tick labels.
        - yticklabels_size : :obj:`int`, default=8
            Font size for y-axis tick labels.
        - shrink : :obj:`float`, default=0.8
            Fraction by which to shrink the colorbar.
        - xlabel_rotation : :obj:`int`, default=0
            Rotation angle for x-axis labels.
        - ylabel_rotation : :obj:`int`, default=0
            Rotation angle for y-axis labels.
        - annot : :obj:`bool`, default=False
            Add values to each cell.
        - annot_kws : :obj:`dict`, default=None,
            Customize the annotations.
        - fmt : :obj:`str`, default=".2g",
            Modify how the annotated vales are presented.
        - linewidths : :obj:`float`, default=0
            Padding between each cell in the plot.
        - borderwidths : :obj:`float`, default=0
            Width of the border around the plot.
        - linecolor : :obj:`str`, default="black"
            Color of the line that seperates each cell.
        - edgecolors : :obj:`str` or :obj:`None`, default=None
            Color of the edges.
        - alpha : :obj:`float` or :obj:`None`, default=None
            Controls transparancy and ranges from 0 (transparant) to 1 (opaque).
        - bbox_inches : :obj:`str` or :obj:`None`, default="tight"
            Alters size of the whitespace in the saved image.
        - cmap : :obj:`str`, :obj:`callable` default="coolwarm"
            Color map for the cells in the plot. For this parameter, you can use premade color palettes or
            create custom ones.
            Below is a list of valid options:

                - Strings to call seaborn's premade palettes.
                - ``seaborn.diverging_palette`` function to generate custom palettes.
                - ``matplotlib.color.LinearSegmentedColormap`` to generate custom palettes.

    Returns
    -------
    `seaborn.heatmap`
        An instance of `seaborn.heatmap`.
    `dict[str, pd.DataFrame]`
        An instance of a pandas DataFrame for each group.

    Note
    ----
    Indices represent "from" and columns represent "to". For instance, the probability at ``df.loc["CAP-1", "CAP-2"]``
    represents the probability from transitioning from CAP-1 to CAP-2.
    """
    assert isinstance(trans_dict,dict), "transition_dict must be in the form dict[str, pd.DataFrame]."

    # Create plot dictionary
    defaults = {"dpi": 300, "figsize": (8, 6), "fontsize": 14, "xticklabels_size": 8, "yticklabels_size": 8,
                "shrink": 0.8, "xlabel_rotation": 0, "ylabel_rotation": 0, "annot": False, "linewidths": 0,
                "linecolor": "black", "cmap": "coolwarm", "fmt": ".2g", "borderwidths": 0, "edgecolors": None,
                "alpha": None, "bbox_inches": "tight", "annot_kws": None}

    plot_dict = _check_kwargs(defaults, **kwargs)

    trans_mat_dict = {}

    for group in trans_dict:
        df = trans_dict[group]
        # Get indices and averaged probabilities
        indices, averaged_probabilities = df.iloc[:,3:].mean().index, df.iloc[:,3:].mean().values
        # Get the maximum CAP
        max_cap = str(max([float(i) for i in indices])).split(".")[0]
        cap_names = [f"CAP-{num}" for num in range(1,int(max_cap) + 1)]
        trans_mat = pd.DataFrame(index=cap_names, columns=cap_names, dtype="float64")

        # Create matrix
        for location, name in enumerate(indices):
            trans_mat.loc[f"CAP-{name.split('.')[0]}", f"CAP-{name.split('.')[1]}"] = averaged_probabilities[location]

        display = _create_display(df=trans_mat, plot_dict=plot_dict, suffix_title=suffix_title, group=group,
                                  call="trans")

        # Store df in dict
        trans_mat_dict[group] = trans_mat

        # Save figure & dataframe
        if output_dir:
            _save_contents(output_dir=output_dir, suffix_title=suffix_title, group=group,
                           curr_dict=trans_mat_dict, plot_dict=plot_dict, save_plots=save_plots,
                           save_df=save_df, display=display, call="trans")

        # Display figures
        plt.show() if show_figs else plt.close()

    if return_df: return trans_mat_dict




import copy, os
from typing import Union, Optional
import numpy as np
from .._utils import _convert_pickle_to_dict, _dicts_to_pickles

def change_dtype(subject_timeseries_list: Union[list[dict[str, dict[str, np.ndarray]]], list[os.PathLike]],
                 dtype: Union[str, np.floating],
                 return_dicts: bool=True,
                 output_dir: Optional[os.PathLike]=None,
                 file_names: Optional[list[str]]=None) -> dict[str, dict[str, dict[str, np.ndarray]]]:

    """
    **Change Dtype of Subject Timeseries**

    Changes the dtypes of each participants numpy array. This function uses the ``.astype()`` method from numpy.
    This function can assist with memory usage. For instance, converting a numpy array from "float64" to "float32"
    can halve the memory required, which is especially useful when performing analyses on larger datasets on a
    local machine.

    Parameters
    ----------
    subject_timeseries_list : :obj:`list[dict[str, dict[str, np.ndarray]]]` or :obj:`list[os.PathLike]`
        A list of dictionaries or pickle files containing the nested subject timeseries dictionary saved by the
        ``TimeSeriesExtractor`` class or a list of nested subject timeseries dictionaries produced by the
        ``TimeSeriesExtractor`` class. The first level of the nested dictionary must consist of the subject ID as a
        string, the second level must consist of the run numbers in the form of "run-#"
        (where # is the corresponding number of the run), and the last level must consist of the timeseries
        (as a numpy array) associated with that run. The structure is as follows:
        ::

            subject_timeseries = {
                    "101": {
                        "run-0": np.array([...]), # 2D array
                        "run-1": np.array([...]), # 2D array
                        "run-2": np.array([...]), # 2D array
                    },
                    "102": {
                        "run-0": np.array([...]), # 2D array
                        "run-1": np.array([...]), # 2D array
                    }
                }

        .. versionchanged:: 0.15 changed from ``subject_timeseries`` to ``subject_timeseries_list``

    dtype : :obj:`bool` or :obj:`np.floating`
        Target data type (e.g "float32" or ``np.float32``) to convert each participant's numpy arrays into.

    return_dicts : :obj:`bool`, default=True
        If True, returns the converted ``subject_timeseries``.

        .. versionchanged:: 0.15 changed from ``return_dict`` to ``return_dicts``

    output_dir : :obj:`os.PathLike` or :obj:`None`, default=None
        Directory to save the converted ``subject_timeseries`` to. Will be saved as a pickle file. The directory will
        be created if it does not exist.

    file_names : :obj:`list[str]` or :obj:`None`, default=None
        A list of names to save the  dictionaries with changed dtypes as. The assignment of file names to dictionaries
        depends on the index position (a file name in the 0th position will be the file name for the dictionary in the
        0th position of ``subject_timeseries_list``). If no names are provided and ``output_dir`` is specified,
        default names will be used.

        .. versionchanged:: 0.15.0 from ``file_name`` to ``file_names``

    Returns
    -------
    `dict[str, dict[str, dict[str, np.ndarray]]]`

    Warning
    -------
    While this function allows conversion to any valid numpy dtype, it is recommended to use floating-point dtypes.
    Reducing the dtype could introduce rounding errors that may lower the precision of subsequent analyses as decimal
    digits are reduced when lower dtypes are requested. Thus, the lowest recommended floating-point dtype would be
    "float32", since it allows for memory usage reduction while limiting rounding errors that could significantly
    alter calculations.
    """
    assert isinstance(subject_timeseries_list, list), "`subject_timeseries_list` must be a list."
    changed_dtype_dicts = {}

    for indx, curr_dict in enumerate(subject_timeseries_list):
        if isinstance(curr_dict, str) and curr_dict.endswith(".pkl"):
            curr_dict = _convert_pickle_to_dict(curr_dict)
        else:
            curr_dict =  copy.deepcopy(curr_dict)
        for subj_id in curr_dict:
            for run in curr_dict[subj_id]:
                curr_dict[subj_id][run] = curr_dict[subj_id][run].astype(dtype)
        changed_dtype_dicts[f"dict_{indx}"] = curr_dict

    if output_dir:
        _dicts_to_pickles(output_dir=output_dir, dict_list=changed_dtype_dicts, file_names=file_names, call="change_dtype")

    if return_dicts: return changed_dtype_dicts



def _check_kwargs(defaults, **kwargs):

    plot_dict = defaults.copy()
    plot_dict.update({k: v for k, v in kwargs.items() if k in plot_dict})

    if kwargs:
        invalid_kwargs = {k: v for k, v in kwargs.items() if k not in plot_dict}
        if invalid_kwargs:
            print(f"Invalid kwargs arguments used and will be ignored {invalid_kwargs}.")

    return plot_dict



from .cap import CAP
from .change_dtype import change_dtype
from .merge import merge_dicts
from .standardize import standardize
from .transition_matrix import transition_matrix

__all__=["CAP", "change_dtype", "merge_dicts", "standardize", "transition_matrix"]
